# type: ignore
# We can either ignore each line individually, or ignore the entire file and wait until mypy can ignore
# specific errors per-file.
import inspect
import sys
from types import FunctionType
from typing import Any, Dict, List, Tuple, Union, _GenericAlias

from classiq_interface.generator.synthesis_metrics import WireRole as Role

from classiq.exceptions import ClassiqQFuncError
from classiq.quantum_register import AuxQReg, QReg, QSFixed, ZeroQReg

if sys.version_info >= (3, 9):
    from types import GenericAlias
else:
    GenericAlias = _GenericAlias

GenericAliasUnion = Union[GenericAlias, _GenericAlias]


class AnnotationParser:
    def __init__(self, func: FunctionType):
        self._func = func

        self.output_types: Dict[str, GenericAlias] = {}

    def parse(self) -> None:
        annotations = self._func.__annotations__.copy()

        # Todo: remove this `if` after introducing `Inplace`
        if "return" not in annotations:
            raise ClassiqQFuncError("Return value annotations not found")

        self.output_values = self._unpack_output_values(annotations.pop("return"))
        self.input_names, self.input_values = self._unpack_input_values(annotations)

        self._validate()

    def _validate(self) -> None:
        self._validate_type_hints()
        self._validate_qubit_amount()
        self._validate_io_length()
        self._validate_io_correlation()

    def _validate_type_hints(self) -> None:
        # Validate type of type-hints
        if not all(
            map(
                self.is_valid_generic_alias_of_qreg,
                self.input_values + self.output_values,
            )
        ):
            raise ClassiqQFuncError("Invalid GenericAlias convection")

    def _validate_qubit_amount(self) -> None:
        # Validate qubit amount
        if sum(i.size for i in self.output_values) != sum(
            i.size for i in self.input_values
        ):
            raise ClassiqQFuncError(
                "Input and output values have different amounts of qubits"
            )

    # Todo: Remove this validation by introducing better heuristics
    #   Or after introducing Inplace
    def _validate_io_length(self) -> None:
        # Validate amount of inputs and outputs
        if len(self.input_values) != len(self.output_values):
            raise ClassiqQFuncError(
                "Inputs and outputs must have the same number of QRegs"
            )

    def _validate_io_correlation(self) -> None:
        # Validate correspondence between inputs and outputs
        for input_name, input_type, output_type in zip(
            self.input_names, self.input_values, self.output_values
        ):
            # Is arithmetic QReg
            if issubclass(input_type.__origin__, QSFixed):
                if input_type != output_type:
                    raise ClassiqQFuncError(
                        f"Arithmetic QReg must be of the same type and size in both the input and the output. Got {input_type} and {output_type}"
                    )

            # Is Auxillary QReg
            if issubclass(input_type.__origin__, AuxQReg):
                if input_type != output_type:
                    raise ClassiqQFuncError(
                        f"Auxillary QReg must be of the same type and size in both the input and the output. Got {input_type} and {output_type}"
                    )

            # Is Zero QReg
            if input_type.__origin__ is ZeroQReg:
                if output_type.__origin__ is QReg or issubclass(
                    output_type.__origin__, QSFixed
                ):
                    self.output_types[input_name] = output_type
                else:
                    raise ClassiqQFuncError(
                        "Invalid output type. Any ZeroQReg in the input must have a corresponding QReg in the output"
                    )

    @classmethod
    def _unpack_output_values(
        cls, output_value_type_hint: Any
    ) -> Tuple[GenericAlias, ...]:
        # Handle QReg type hints
        if cls.is_subclass_qreg(output_value_type_hint):
            return (cls.to_generic_alias(output_value_type_hint),)

        # Supporting both `typing._GenericAlias` and `types.GenericAlias`
        if not cls.is_instance_generic_alias(output_value_type_hint):
            raise ClassiqQFuncError(
                "Output value type hint must be either a single QReg, `typing.Tuple[QReg, ...]` or, for python>=3.9, `tuple[QReg, ...]`"
            )

        # Allowing only a tuple of outputs:
        if not cls.is_tuple_generic_alias(output_value_type_hint):
            raise ClassiqQFuncError(
                "Output value type hint must be either Tuple[QReg, ...] or tuple[QReg, ...]"
            )

        # This line may raise ClassiqQFuncError
        return tuple(map(cls.to_generic_alias, output_value_type_hint.__args__))

    def _unpack_input_values(
        self, annotations: Dict[str, Any]
    ) -> Tuple[List[str], Tuple[GenericAlias, ...]]:
        input_names = list(annotations.keys())
        input_values = tuple(map(self.to_generic_alias, annotations.values()))
        return input_names, input_values

    @staticmethod
    def to_generic_alias(obj: Any) -> GenericAlias:
        # Handle GenericAlias
        if isinstance(obj, GenericAlias):
            return obj
        # Handle _GenericAlias, for python>3.9, i.e. when GenericAlias != _GenericAlias
        if isinstance(obj, _GenericAlias):
            return GenericAlias(obj.__origin__, obj.__args__)
        # Handle a single QReg (not GenericAlias of QReg)
        elif inspect.isclass(obj) and issubclass(obj, QReg):
            return GenericAlias(obj, tuple())

        raise ClassiqQFuncError(f"Invalid type hint object: {obj.__class__.__name__}")

    @staticmethod
    def is_instance_generic_alias(obj: Any) -> bool:
        return isinstance(obj, (GenericAlias, _GenericAlias))

    @classmethod
    def is_subclass_qreg(cls, obj: Any) -> bool:
        if inspect.isclass(obj):
            return issubclass(obj, QReg)
        elif cls.is_instance_generic_alias(obj):
            return issubclass(obj.__origin__, QReg)
        return False

    @staticmethod
    def is_tuple_generic_alias(obj: GenericAliasUnion) -> bool:
        return obj.__origin__.__name__.lower() == "tuple"

    @staticmethod
    def is_valid_generic_alias_of_qreg(obj: GenericAlias) -> bool:
        return (
            isinstance(obj, GenericAlias)
            and issubclass(obj.__origin__, QReg)
            and type(obj.__args__) is tuple
            # The tests below this comment force annotations with size
            and len(obj.__args__) == 1
            and type(obj.__args__[0]) is int
        )


def get_annotation_role(annotation: GenericAlias, default_role: Role) -> Role:
    """
    Note: this function cannot distinguish between inputs and outputs.
        Thus, for inputs, all 3 options are valid
        However, for outputs:
            a) we don't expect to get ZERO
            b) We treat INPUT as OUTPUT
    """
    ret = None

    if getattr(annotation, "role", None) is not None:
        ret = annotation.role
    if getattr(annotation.__origin__, "role", None) is not None:
        ret = annotation.role

    if issubclass(annotation.__origin__, QReg) and not issubclass(
        annotation.__origin__, ZeroQReg
    ):
        ret = default_role

    if issubclass(annotation.__origin__, ZeroQReg) and not issubclass(
        annotation.__origin__, AuxQReg
    ):
        ret = Role.ZERO

    if issubclass(annotation.__origin__, AuxQReg):
        ret = Role.AUXILIARY

    # Didn't match anything so far
    if ret is None:
        raise ClassiqQFuncError("Invalid annotation role")

    if default_role == Role.INPUT and ret == Role.OUTPUT:
        raise ClassiqQFuncError("input should not have Role.OUTPUT")

    if default_role == Role.OUTPUT and ret in (Role.ZERO, Role.INPUT):
        raise ClassiqQFuncError("output should not have Role.ZERO / Role.INPUT")

    return ret



"""Configuration of the SDK module."""
import os
import pathlib
from typing import List, Optional, Union

import configargparse  # type: ignore[import]
import pydantic

from classiq.exceptions import ClassiqValueError


class WebsocketURI(pydantic.AnyUrl):
    allowed_schemes = {"ws", "wss"}


class Configuration(pydantic.BaseModel):
    """Object containing configuration options (see description in fields)."""

    host: pydantic.AnyHttpUrl = pydantic.Field(..., description="Classiq backend URI.")
    ws_uri: WebsocketURI = pydantic.Field(
        default=None,
        description="Web socket URI, derived automatically from the host field.",
    )
    should_check_host: bool = pydantic.Field(
        default=True, description="Should check backend URI and version."
    )

    @pydantic.validator("ws_uri", always=True)
    def validate_ws_uri(cls, v, values) -> WebsocketURI:
        if v is not None:
            return v

        host = values.get("host")
        if host is None:
            raise ClassiqValueError("Can not init ws_uri with out a valid host")

        scheme = "wss" if host.scheme == "https" else "ws"
        port = f":{host.port}" if host.port else ""
        return pydantic.parse_obj_as(WebsocketURI, f"{scheme}://{host.host}{port}")


_DEFAULT_CONFIG_FILES = [str(pathlib.Path("classiq", "config.ini"))]
if os.name == "posix":
    # Unix convensions:
    #   System-wide configuration rests in "/etc"
    #       either as "/etc/program_name.conf" or as "/etc/program_name/some_name"
    #   User-wide configuration rests in "~/.config"
    # Order matters - System-wide is most general, than user-wide,
    #   and than folder-specific configration
    _DEFAULT_CONFIG_FILES = [
        "/etc/classiq/config.ini",
        "/etc/classiq.conf",
        "~/.config/classiq/config.ini",
        "~/.config/classiq.conf",
    ] + _DEFAULT_CONFIG_FILES


def init(args: Optional[Union[str, List[str]]] = None) -> Configuration:
    """Initialize the configuration object.

    Args:
        args (): Non-default arguments.

    Returns:
        Initialized configuration object.
    """
    arg_parser = configargparse.ArgParser(default_config_files=_DEFAULT_CONFIG_FILES)

    arg_parser.add_argument(
        "--classiq-config-file",
        is_config_file=True,
        help="Configuration file path",
        env_var="CLASSIQ_CONFIG_FILE",
    )
    arg_parser.add_argument(
        "--classiq-host",
        help="The URL of Classiq's backend host",
        env_var="CLASSIQ_HOST",
        default="https://classiquantum.com",
    )
    arg_parser.add_argument(
        "--classiq-skip-check-host",
        dest="classiq_skip_check_host",
        help="Should skip classiq host and version",
        env_var="CLASSIQ_SKIP_CHECK_HOST",
        action="store_true",
    )

    parsed_args, _ = arg_parser.parse_known_args(args=args)
    return Configuration(
        host=parsed_args.classiq_host,
        should_check_host=not parsed_args.classiq_skip_check_host,
    )



import asyncio
from typing import Any, Dict, List, Optional

from pyomo.core import ConcreteModel

from classiq_interface import status
from classiq_interface.backend.backend_preferences import BackendPreferences
from classiq_interface.chemistry import operator
from classiq_interface.combinatorial_optimization import (
    model_serializer,
    optimization_problem,
    sense,
)
from classiq_interface.combinatorial_optimization.encoding_types import EncodingType
from classiq_interface.combinatorial_optimization.preferences import QAOAPreferences
from classiq_interface.executor.execution_preferences import OptimizerPreferences
from classiq_interface.executor.result import ExecutionData, ExecutionStatus
from classiq_interface.generator import result as generator_result
from classiq_interface.generator.result import GeneratedCircuit

from classiq._internals.api_wrapper import ApiWrapper
from classiq._internals.type_validation import validate_type
from classiq.exceptions import (
    ClassiqError,
    ClassiqExecutionError,
    ClassiqGenerationError,
)


class CombinatorialOptimization:
    def __init__(
        self,
        model: ConcreteModel,
        qaoa_preferences: Optional[QAOAPreferences] = None,
        optimizer_preferences: Optional[OptimizerPreferences] = None,
        backend_preferences: Optional[BackendPreferences] = None,
        encoding_type: Optional[EncodingType] = None,
    ):
        self.is_maximization = sense.is_maximization(model)
        self._serialized_model = model_serializer.to_json(model, return_dict=True)

        arguments: Dict[str, Any] = {
            "serialized_model": self._serialized_model,
            "encoding_type": encoding_type,
        }
        if qaoa_preferences is not None:
            arguments["qaoa_preferences"] = qaoa_preferences
        if optimizer_preferences is not None:
            arguments["optimizer_preferences"] = optimizer_preferences
        if backend_preferences is not None:
            arguments["backend_preferences"] = backend_preferences

        self._problem = optimization_problem.OptimizationProblem(**arguments)

    def generate(self) -> GeneratedCircuit:
        return asyncio.run(self.generate_async())

    async def generate_async(self) -> GeneratedCircuit:
        """Async version of `generate`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_generate_task(
            problem=self._problem
        )

        if result.status != generator_result.GenerationStatus.SUCCESS:
            raise ClassiqGenerationError(f"Solving failed: {result.details}")

        return validate_type(
            obj=result.details,
            expected_type=GeneratedCircuit,
            operation="Solving",
            exception_type=ClassiqGenerationError,
        )

    def solve(self) -> ExecutionData:
        return asyncio.run(self.solve_async())

    async def solve_async(self) -> ExecutionData:
        """Async version of `solve`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_solve_task(
            problem=self._problem
        )

        if result.status != ExecutionStatus.SUCCESS:
            raise ClassiqExecutionError(f"Solving failed: {result.details}")
        return result.details

    def solve_classically(self) -> ExecutionData:
        return asyncio.run(self.solve_classically_async())

    async def solve_classically_async(self) -> ExecutionData:
        """Async version of `solve`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_solve_classically_task(
            problem=self._problem
        )

        if result.status != ExecutionStatus.SUCCESS:
            raise ClassiqExecutionError(f"Solving failed: {result.details}")

        return result.details

    def get_operator(self) -> operator.PauliOperator:
        return asyncio.run(self.get_operator_async())

    async def get_operator_async(self) -> operator.PauliOperator:
        """Async version of `get_operator`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_operator_task(
            problem=self._problem
        )

        if result.status != operator.OperatorStatus.SUCCESS:
            raise ClassiqError(f"Get operator failed: {result.details}")
        return validate_type(
            obj=result.details,
            expected_type=operator.PauliOperator,
            operation="Get operator",
            exception_type=ClassiqGenerationError,
        )

    def get_objective(self) -> str:
        return asyncio.run(self.get_objective_async())

    async def get_objective_async(self) -> str:
        """Async version of `get_operator`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_objective_task(
            problem=self._problem
        )

        if result.status != status.Status.SUCCESS:
            raise ClassiqError(f"Get objective failed: {result.details}")

        return result.details

    def get_initial_point(self) -> List[float]:
        return asyncio.run(self.get_initial_point_async())

    async def get_initial_point_async(self) -> List[float]:
        """Async version of `get_initial_point`"""
        wrapper = ApiWrapper()
        result = await wrapper.call_combinatorial_optimization_initial_point_task(
            problem=self._problem
        )

        if result.status != status.Status.SUCCESS:
            raise ClassiqError(f"Get inital point failed: {result.details}")

        return result.details



import functools
import json
import logging
from typing import Dict, Optional

import httpx
import websockets.client
import websockets.exceptions
import websockets.typing

from classiq_interface.analyzer import analysis_params, result as analysis_result
from classiq_interface.chemistry import ground_state_problem, operator
from classiq_interface.combinatorial_optimization import (
    optimization_problem,
    result as opt_result,
)
from classiq_interface.executor import execution_request, result as execute_result
from classiq_interface.generator import result as generator_result
from classiq_interface.generator.model import Model
from classiq_interface.server import routes

from classiq._internals import client
from classiq.exceptions import ClassiqExpiredTokenError, ClassiqValueError

_FAIL_FAST_INDICATOR = "{"


def _retry_websocket_on_exception(func):
    # TODO fix this workaround
    @functools.wraps(func)
    async def wrapped(self: "ApiWrapper", *args, **kwargs):
        try:
            return await func(self, *args, **kwargs)
        except websockets.exceptions.InvalidStatusCode as exc:
            # If the token is missing or invalid, the handshake fails
            # with a status code of 403. In this case, we refresh the token
            # and try again.
            if exc.status_code == httpx.codes.FORBIDDEN.value:
                await self._client.update_expired_access_token()
                return await func(self, *args, **kwargs)
            raise

    return wrapped


def _decode_message(msg: websockets.typing.Data) -> str:
    return msg.decode() if isinstance(msg, bytes) else msg


async def _get_return_value_from_websocket(
    websocket: websockets.client.WebSocketClientProtocol,
) -> str:
    is_first = True
    msg: str = _decode_message(await websocket.recv())

    while msg:
        logging.info(msg)

        # HACK: This is meant to handle validation errors. It is a duplicate
        # of a similar hack implemented in the VS code extension
        if is_first:
            is_first = False
            if msg.startswith(_FAIL_FAST_INDICATOR):
                return msg

        msg = _decode_message(await websocket.recv())

    return _decode_message(await websocket.recv())


class ApiWrapper:
    def __init__(self) -> None:
        self._client = client.client()

    async def _call_task(self, http_method: str, url: str, body: Optional[Dict] = None):
        try:
            res = await self._client.call_api(
                http_method=http_method, url=url, body=body
            )
        except ClassiqExpiredTokenError:
            await self._client.update_expired_access_token()
            res = await self._client.call_api(
                http_method=http_method, url=url, body=body
            )

        if not isinstance(res, dict):
            raise ClassiqValueError(f"Unexpected returned value: {res}")

        return res

    @_retry_websocket_on_exception
    async def call_generation_task(
        self, model: Model
    ) -> generator_result.GenerationResult:
        async with self._client.establish_websocket_connection(
            path="/api/v1/ws/generate"
        ) as websocket:
            await websocket.send(model.json())
            res = await _get_return_value_from_websocket(websocket)
            return generator_result.GenerationResult.parse_raw(res)

    async def call_execute_task(
        self, request: execution_request.ExecutionRequest
    ) -> execute_result.ExecutionResult:
        data = await self._call_task(
            http_method="post",
            url=routes.EXECUTE_TASKS_FULL_PATH,
            body=json.loads(
                request.json()
            ),  # TODO: request.dict() doesn't serialize complex class
        )

        return execute_result.ExecutionResult.parse_obj(data)

    async def call_analysis_task(
        self, params: analysis_params.AnalysisParams
    ) -> analysis_result.AnalysisResult:
        data = await self._call_task(
            http_method="post",
            url=routes.ANALYZER_FULL_PATH,
            body=params.dict(),
        )

        return analysis_result.AnalysisResult.parse_obj(data)

    async def call_combinatorial_optimization_generate_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> generator_result.GenerationResult:
        data = await self._call_task(
            http_method="post",
            url=routes.COMBINATORIAL_OPTIMIZATION_GENERATE_FULL_PATH,
            body=problem.dict(),
        )

        return generator_result.GenerationResult.parse_obj(data)

    @_retry_websocket_on_exception
    async def call_combinatorial_optimization_solve_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> execute_result.ExecutionResult:
        async with self._client.establish_websocket_connection(
            path=routes.COMBINATORIAL_OPTIMIZATION_WS_SOLVE_FULL_PATH, ping_interval=20
        ) as websocket:
            await websocket.send(problem.json())
            res = await _get_return_value_from_websocket(websocket)
            return execute_result.ExecutionResult.parse_raw(res)

    async def call_combinatorial_optimization_solve_classically_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> execute_result.ExecutionResult:
        data = await self._call_task(
            http_method="post",
            url=routes.COMBINATORIAL_OPTIMIZATION_SOLVE_CLASSICALLY_FULL_PATH,
            body=problem.dict(),
        )

        return execute_result.ExecutionResult.parse_obj(data)

    async def call_combinatorial_optimization_operator_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> operator.OperatorResult:
        data = await self._call_task(
            http_method="post",
            url=routes.COMBINATORIAL_OPTIMIZATION_OPERATOR_FULL_PATH,
            body=problem.dict(),
        )

        return operator.OperatorResult.parse_obj(data)

    async def call_combinatorial_optimization_objective_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> opt_result.PyomoObjectResult:
        data = await self._call_task(
            http_method="post",
            url=routes.COMBINATORIAL_OPTIMIZATION_OBJECTIVE_FULL_PATH,
            body=problem.dict(),
        )

        return opt_result.PyomoObjectResult.parse_obj(data)

    async def call_combinatorial_optimization_initial_point_task(
        self, problem: optimization_problem.OptimizationProblem
    ) -> opt_result.AnglesResult:
        data = await self._call_task(
            http_method="post",
            url=routes.COMBINATORIAL_OPTIMIZATION_INITIAL_POINT_FULL_PATH,
            body=problem.dict(),
        )

        return opt_result.AnglesResult.parse_obj(data)

    async def call_generate_hamiltonian_task(
        self, problem: ground_state_problem.GroundStateProblem
    ) -> operator.OperatorResult:
        data = await self._call_task(
            http_method="post",
            url=routes.GENERATE_CHEMISTRY_HAMILTONIAN_FULL_PATH,
            body=problem.dict(),
        )

        return operator.OperatorResult.parse_obj(data)



import sys
from types import FunctionType
from typing import Callable, Dict, List, Optional, Tuple, cast

from classiq_interface.generator.functions import (
    FunctionData,
    FunctionImplementation,
    Register,
)
from classiq_interface.generator.synthesis_metrics import WireRole as Role

# This line is ignored because the entire annotation_parser module is ignored by mypy
from classiq.quantum_functions.annotation_parser import (  # type: ignore[attr-defined]
    AnnotationParser,
    get_annotation_role,
)
from classiq.quantum_functions.function_parser import (
    FunctionParser,
    convert_callable_to_function,
)

# isort: split
from typing import _GenericAlias  # type: ignore[attr-defined]

if sys.version_info >= (3, 9):
    from types import GenericAlias
else:
    GenericAlias = _GenericAlias


class QuantumFunction:
    def __init__(self) -> None:
        self._function_data: Optional[FunctionData] = None

    @staticmethod
    def _generate_single_register(
        first_qubit: int, name: str, obj: GenericAlias  # type: ignore[valid-type]
    ) -> Register:
        qreg_size = obj.size  # type: ignore[attr-defined]

        qubits = tuple(range(first_qubit, first_qubit + qreg_size))

        return Register(
            name=name,
            qubits=qubits,
        )

    @classmethod
    def _generate_registers(
        cls, input_names, input_values, output_values
    ) -> Dict[Role, Tuple[Register, ...]]:
        registers: Dict[Role, List[Register]] = {k: list() for k in Role}

        qubit_counter = 0
        for input_name, input_annotation in zip(input_names, input_values):
            role = get_annotation_role(
                annotation=input_annotation, default_role=Role.INPUT
            )

            registers[role].append(
                cls._generate_single_register(
                    first_qubit=qubit_counter,
                    name=input_name,
                    obj=input_annotation,
                )
            )
            qubit_counter += input_annotation.size

        qubit_counter = 0
        for input_name, output_annotation in zip(input_names, output_values):
            role = get_annotation_role(
                annotation=output_annotation, default_role=Role.OUTPUT
            )
            if role == Role.AUXILIARY:
                qubit_counter += output_annotation.size
                continue

            registers[role].append(
                cls._generate_single_register(
                    first_qubit=qubit_counter,
                    name=input_name,
                    obj=output_annotation,
                )
            )
            qubit_counter += output_annotation.size

        return {k: tuple(v) for k, v in registers.items()}

    @classmethod
    def _create_implementation_from_function(
        cls, func: FunctionType, func_name: str
    ) -> FunctionImplementation:
        # Return value
        fp = FunctionParser(func)
        serialized_circuit = fp.extract_function_output()

        # Annotations
        ap = AnnotationParser(func)
        ap.parse()

        registers: Dict[Role, Tuple[Register, ...]] = cls._generate_registers(
            input_names=ap.input_names,
            input_values=ap.input_values,
            output_values=ap.output_values,
        )

        implementation = FunctionImplementation(
            name=func_name,
            serialized_circuit=serialized_circuit,
            input_registers=registers[Role.INPUT],
            output_registers=registers[Role.OUTPUT],
            zero_input_registers=registers[Role.ZERO],
            auxiliary_registers=registers[Role.AUXILIARY],
        )
        return implementation

    def add_implementation(self, func: Callable) -> "QuantumFunction":
        func, func_name = convert_callable_to_function(func=func)

        implementation = self._create_implementation_from_function(
            func=func, func_name=func_name
        )

        if self._function_data is None:
            self._function_data = FunctionData(
                name=func_name,
                implementations=(implementation,),
            )
        else:
            new_implementations = cast(
                Tuple[FunctionImplementation], self._function_data.implementations
            ) + (implementation,)
            self._function_data.implementations = (
                self._function_data.validate_implementations(new_implementations)
            )

        return self

    @property
    def function_data(self):
        return self._function_data



"""ModelDesigner module, implementing facilities for designing models and generating circuits using Classiq platform."""
from __future__ import annotations

import asyncio
import distutils.spawn
import logging
import tempfile
from contextlib import nullcontext
from typing import IO, AnyStr, ContextManager, List, Optional, Type, Union

from classiq_interface.generator import function_call, result
from classiq_interface.generator.model import Constraints, Model, Preferences

from classiq._internals import api_wrapper
from classiq.exceptions import ClassiqFileNotFoundError, ClassiqGenerationError
from classiq.model_designer import function_handler, wire
from classiq.quantum_functions.function_library import FunctionLibrary

_logger = logging.getLogger(__name__)

_SupportedIO = Union[IO, str]

# TODO: Add docstrings for auto generated methods.


def _file_handler(fp: Optional[_SupportedIO], mode: str = "r") -> ContextManager[IO]:
    if fp is None:
        temp_file = tempfile.NamedTemporaryFile(mode, suffix=".qmod", delete=False)
        print(f'Using temporary file: "{temp_file.name}"')
        return temp_file

    if isinstance(fp, str):
        return open(fp, mode)

    return nullcontext(fp)


class ModelDesigner(function_handler.FunctionHandler):
    """Facility to generate circuits, based on the model."""

    def __init__(self, **kwargs) -> None:
        """Init self."""
        super().__init__()
        self._model = Model(**kwargs)

    @property
    def _output_wire_type(self) -> Type[wire.Wire]:
        return wire.Wire

    @property
    def _logic_flow(self) -> List[function_call.FunctionCall]:
        return self._model.logic_flow

    @property
    def constraints(self) -> Constraints:
        """Get the constraints aggregated in self.

        Returns:
            The constraints data.
        """
        return self._model.constraints

    @property
    def preferences(self) -> Preferences:
        """Get the preferences aggregated in self.

        Returns:
            The preferences data.
        """
        return self._model.preferences

    def synthesize(self) -> result.GeneratedCircuit:
        """Generates a circuit, based on the aggregation of requirements in self.

        Returns:
            The results of the generation procedure.
        """
        return asyncio.run(self.synthesize_async())

    async def synthesize_async(self) -> result.GeneratedCircuit:
        """Async version of `generate`
        Generates a circuit, based on the aggregation of requirements in self.

        Returns:
            The results of the generation procedure.
        """
        # TODO: There something distorted with regards to the singleton and the configuration. Also, the need to pass
        #       conf here and not in init is weird.
        wrapper = api_wrapper.ApiWrapper()
        generation_result = await wrapper.call_generation_task(self._model)

        if generation_result.status != result.GenerationStatus.SUCCESS:
            raise ClassiqGenerationError(
                f"Generation failed: {generation_result.details}"
            )

        return generation_result.details

    def include_library(self, library: FunctionLibrary) -> None:
        """Includes a user-defined custom function library.

        Args:
            library (FunctionLibrary): The custom function library.
        """
        super().include_library(library=library)
        self._model.function_library = library.data

    def dumps(self, ignore_warning: bool = False) -> str:
        """Serialize model to a JSON formatted `str`

        Args:
            ignore_warning (bool): Whether to ignore the warning print
        """
        if not ignore_warning:
            _logger.warning(
                "Saving to json is currently unstable since versions may change"
            )

        return self._model.json(exclude_defaults=True, indent=4)

    def dump(
        self, fp: Optional[_SupportedIO] = None, ignore_warning: bool = False
    ) -> None:
        """Serialize model to a JSON formatted stream to `fp` (a `.write()`)-supporting file-like object

        Args:
            fp (IO | str | None): a file-like object
                if None -> a temporaty file will be created
                if str -> this will be treated as the file path
            ignore_warning (bool): Whether to ignore the warning print
        """
        with _file_handler(fp, "w") as f:
            f.write(self.dumps(ignore_warning=ignore_warning))

    @classmethod
    def loads(cls, s: AnyStr) -> ModelDesigner:
        """Deserialize `s`, a JSON formatted `str`, to a ModelDesigner

        Args:
            s (str | bytes): A JSON-formatted `str` | `bytes`
        """
        new_instance = cls()
        new_instance._model = Model.parse_raw(s)
        return new_instance

    @classmethod
    def load(cls, fp: Optional[_SupportedIO]) -> ModelDesigner:
        """Deserialize `fp` (a `.read()`-supporting file-like object) containing a JSON formatted document to a ModelDesigner

        Args:
            fp (IO | str): a file-like object
                if str -> this will be treated as the file path
        """
        with _file_handler(fp, "r") as f:
            return cls.loads(f.read())

    def export_to_vscode(self, ignore_warning: bool = False) -> None:
        """Export the model to a file, and display in VisualStudioCode"""

        if not distutils.spawn.find_executable("code"):
            raise ClassiqFileNotFoundError(
                "Please install VSCode to path\nIn VSCode, press [Ctrl/Command]+Shift+p, and then type \"install 'code' command in PATH\""
            )

        fp = tempfile.NamedTemporaryFile("w", suffix=".qmod", delete=False)
        self.dump(fp, ignore_warning=ignore_warning)
        fp.close()
        distutils.spawn.spawn(["code", fp.name])



import logging
from typing import Optional

import keyring
from keyring.backends import fail

_logger = logging.getLogger(__name__)


class PasswordManager:
    _SERVICE_NAME: str = "classiqTokenService"
    _ACCESS_TOKEN_KEY: str = "classiqTokenAccount"
    _REFRESH_TOKEN_KEY: str = "classiqRefershTokenAccount"

    def __init__(self) -> None:
        self._pm_available = is_password_manager_available()

    @property
    def access_token(self) -> Optional[str]:
        return self._get(key=self._ACCESS_TOKEN_KEY)

    @access_token.setter
    def access_token(self, access_token: Optional[str]) -> None:
        self._set(key=self._ACCESS_TOKEN_KEY, value=access_token)

    @property
    def refresh_token(self) -> Optional[str]:
        return self._get(key=self._REFRESH_TOKEN_KEY)

    @refresh_token.setter
    def refresh_token(self, refresh_token: Optional[str]) -> None:
        self._set(key=self._REFRESH_TOKEN_KEY, value=refresh_token)

    def _get(self, key: str) -> Optional[str]:
        if not self._pm_available:
            _logger.info("Called get with unavailable password manager")
            return None
        return keyring.get_password(service_name=self._SERVICE_NAME, username=key)

    def _set(self, key: str, value: Optional[str]) -> None:
        if not self._pm_available:
            _logger.info("Called set with unavailable password manager")
            return

        if value is None:
            self._clear(key)
            return
        keyring.set_password(
            service_name=self._SERVICE_NAME,
            username=key,
            password=value,
        )

    def _clear(self, key: str) -> None:
        if not self._pm_available:
            _logger.info("Called clear with unavailable password manager")
            return

        keyring.delete_password(
            service_name=self._SERVICE_NAME,
            username=key,
        )


class DummyPasswordManager(PasswordManager):
    def _get(self, key: str) -> Optional[str]:
        return None

    def _set(self, key: str, value: Optional[str]) -> None:
        return

    def _clear(self, key: str) -> None:
        return


PASSWORD_MANAGER_AVAILABLE = None


def is_password_manager_available() -> bool:
    global PASSWORD_MANAGER_AVAILABLE
    if PASSWORD_MANAGER_AVAILABLE is None:
        PASSWORD_MANAGER_AVAILABLE = not isinstance(keyring.get_keyring(), fail.Keyring)

    return PASSWORD_MANAGER_AVAILABLE



from typing import Dict, Iterable, List, Set, Type

from classiq_interface.generator import function_call
from classiq_interface.generator.functions import FunctionData

from classiq.exceptions import ClassiqWiringError
from classiq.model_designer import function_handler, wire


class CompositeFunctionInputWire(wire.Wire):
    def __init__(self, input_name: str) -> None:
        super().__init__()
        self._initialize_wire_name(wire_name=input_name)

    @property
    def is_ended(self) -> bool:
        return self._end_call is not None


class CompositeFunctionOutputWire(wire.Wire):
    @property
    def is_ended(self) -> bool:
        return self.is_started and self._start_name in self._start_call.outputs  # type: ignore[union-attr]


class CompositeFunctionGenerator(function_handler.FunctionHandler):
    def __init__(self, function_name: str) -> None:
        super().__init__()
        self._name = function_name
        self._logic_flow_list: List[function_call.FunctionCall] = list()
        self._input_names: Set[str] = set()

    @property
    def _logic_flow(self) -> List[function_call.FunctionCall]:
        return self._logic_flow_list

    @property
    def _output_wire_type(self) -> Type[wire.Wire]:
        return CompositeFunctionOutputWire

    def create_inputs(
        self, input_names: Iterable[str]
    ) -> Dict[str, CompositeFunctionInputWire]:
        input_names = list(input_names)
        if len(input_names) != len(set(input_names)):
            raise ClassiqWiringError("Cannot create multiple inputs with the same name")

        wire_dict = {
            name: CompositeFunctionInputWire(input_name=name) for name in input_names
        }
        self._update_generated_wires(wires=wire_dict.values())

        return wire_dict

    def set_outputs(self, outputs: Dict[str, CompositeFunctionOutputWire]) -> None:
        self._verify_legal_wires(wires=outputs.values())

        for output_name, output_wire in outputs.items():
            if isinstance(output_wire, CompositeFunctionInputWire):
                raise ClassiqWiringError(
                    f"Can't connect input directly to output {output_name}"
                )
            output_wire.set_as_output(output_name=output_name)

    def to_function_data(self) -> FunctionData:
        return FunctionData(name=self._name, logic_flow=self._logic_flow)



from classiq.applications.combinatorial_optimization import (  # noqa: F401
    CombinatorialOptimization,
)
from classiq.applications.ground_state_problem import (  # noqa: F401
    generate_hamiltonian,
    generate_hamiltonian_async,
)



import sys
from types import FunctionType, MethodType
from typing import Any, Callable, Tuple, _GenericAlias  # type: ignore[attr-defined]

if sys.version_info >= (3, 9):
    from types import GenericAlias
else:
    GenericAlias = _GenericAlias


from classiq.exceptions import ClassiqQFuncError


class FunctionParser:
    def __init__(self, func: FunctionType):
        self._func = func

    @staticmethod
    def _extract_function_output_by_execution(func: FunctionType) -> Any:
        # Todo: Parse the type (annotations) of the arguments, and remove only those inheriting from QReg
        if func.__code__.co_kwonlyargcount:
            raise ClassiqQFuncError("kw only args are not supported")

        arg_count = func.__code__.co_argcount - ("self" in func.__code__.co_varnames)
        nones = [None] * arg_count
        return func(*nones)

    def extract_function_output(self) -> Any:
        output = self._extract_function_output_by_execution(func=self._func)

        self._validate_function_output(output)

        return output

    @staticmethod
    def _validate_function_output(output: Any) -> None:
        # Todo: validate QASM

        if type(output) is not str:
            raise ClassiqQFuncError(
                "Invalid output. Please return a string of OpenQASM2.0."
            )


def _convert_class_to_function(cls: type) -> Tuple[FunctionType, str]:
    # Create instance
    try:
        inst = cls()
    except TypeError as exc:
        raise ClassiqQFuncError("Unable to initialize class") from exc

    return inst.__call__, inst.__class__.__name__


def convert_callable_to_function(func: Callable) -> Tuple[FunctionType, str]:
    # There's a story to be told here:
    # Functions vs Methods.
    # It's a centuries old fight, which won't end soon..
    # Functions, being functions, always have a way of telling us about their functionality
    #   This is done using `func.__code__`
    # Methods, being the heigher form of functions, are too proud to have a `__code__` attribute.
    #   Thus, `"__code__" in method` is false
    #   However, `method.__code__` exists
    #   Since it is inherited from the class that initialized the instance who owns the method
    # Thus, we reach the important conclusion:
    # Methods are shy functions. They are functions, but they don't like telling us that they're functions.
    if not callable(func):
        raise ClassiqQFuncError("Invalid callable given.")

    if isinstance(func, (FunctionType, MethodType)):
        return func, func.__name__  # type: ignore[return-value]

    # Assuming `func` is an instance of some class
    return _convert_class_to_function(func)  # type: ignore[arg-type]



from typing import Any, Tuple, Type, TypeVar, Union, overload

T = TypeVar("T")
U = TypeVar("U")


# Overloads are provided for the common case and for the general case separately since python has poor support for
# variadic type variables.


@overload
def validate_type(
    obj: Any, expected_type: Type[T], operation: str, exception_type: Type[Exception]
) -> T:
    ...


@overload
def validate_type(
    obj: Any,
    expected_type: Tuple[Type[T], Type[U]],
    operation: str,
    exception_type: Type[Exception],
) -> Union[T, U]:
    ...


@overload
def validate_type(
    obj: Any,
    expected_type: Tuple[Type[T], ...],
    operation: str,
    exception_type: Type[Exception],
) -> Any:
    ...


def validate_type(
    obj: Any,
    expected_type: Union[Tuple[type, ...], type],
    operation: str,
    exception_type: Type[Exception],
) -> Any:
    operation = operation[0].upper() + operation[1:]
    if not isinstance(obj, expected_type):
        expected_types: Tuple[type, ...]
        if isinstance(expected_type, type):
            expected_types = (expected_type,)
        else:
            expected_types = expected_type
        type_str = " or ".join(t.__name__ for t in expected_types)
        raise exception_type(
            f"{operation} error: Got object of type {type(obj).__name__} and not {type_str}"
        )
    return obj



import abc
import collections.abc
import functools
from typing import Collection, Dict, Iterable, List, Optional, Set, Type, Union

from classiq_interface.generator import (
    function_call,
    function_param_list,
    function_params,
)
from classiq_interface.generator.functions import FunctionLibraryData
from classiq_interface.generator.user_defined_function_params import CustomFunction

from classiq.exceptions import ClassiqValueError, ClassiqWiringError
from classiq.model_designer.wire import Wire
from classiq.quantum_functions.function_library import FunctionLibrary
from classiq.quantum_register import QReg as QuantumRegister

QregOrWire = Union[QuantumRegister, Wire]
WireOrWires = Union[Wire, Iterable[Wire]]
SupportedIOUnion = Union[
    Dict[str, QregOrWire],
    Collection[QuantumRegister],
    QuantumRegister,
]


class FunctionHandler(abc.ABC):
    def __init__(self) -> None:
        self._generated_wires: Set[Wire] = set()
        self._function_library: Optional[FunctionLibrary] = None

    def _verify_legal_wires(self, wires: WireOrWires) -> None:
        if isinstance(wires, Wire):
            wires = [wires]
        if not all(wire in self._generated_wires for wire in wires):
            raise ClassiqWiringError("Wire does not belong to this generator")

    def _update_generated_wires(self, wires: WireOrWires) -> None:
        if isinstance(wires, Wire):
            wires = [wires]
        self._generated_wires.update(wires)

    def apply(
        self,
        function_name: str,
        in_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = None,
        out_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = None,
    ) -> Dict[str, Wire]:
        if self._function_library is None:
            raise ClassiqValueError("Cannot apply function without a function library")

        params = self._function_library.get_function(function_name=function_name)
        return self._function_call_handler(
            function=CustomFunction.__name__,
            params=params,
            in_wires=in_wires,
            out_wires=out_wires,
        )

    def _function_call_handler(
        self,
        function: str,
        params: function_params.FunctionParams,
        in_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = None,
        out_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = None,
        is_inverse: bool = False,
    ) -> Dict[str, Wire]:
        if function != type(params).__name__:
            raise ClassiqValueError(
                "The FunctionParams type does not match function name"
            )

        if isinstance(params, CustomFunction):
            FunctionLibraryData.validate_function_in_library(
                library=self._function_library.data if self._function_library else None,
                function_params=params,
                error_handler=ClassiqValueError,
            )

        call = function_call.FunctionCall(
            function=function, function_params=params, is_inverse=is_inverse
        )

        if in_wires is not None:
            self._connect_in_wires(call=call, in_wires=in_wires)

        self._logic_flow.append(call)

        return self._connect_out_wires(
            call=call,
            out_wires=out_wires or {},
        )

    def _connect_in_wires(
        self,
        call: function_call.FunctionCall,
        in_wires: SupportedIOUnion,
    ) -> None:
        if isinstance(in_wires, dict):
            self._connect_named_in_wires(call=call, in_wires=in_wires)
        elif isinstance(in_wires, QuantumRegister):
            self._connect_unnamed_in_quantum_registers(
                call=call, quantum_registers=[in_wires]
            )
        elif isinstance(in_wires, collections.abc.Collection):
            self._connect_unnamed_in_quantum_registers(
                # mypy doesn't recognize that `dict` wouldn't reach this point
                call=call,
                quantum_registers=in_wires,  # type: ignore[arg-type]
            )
        else:
            raise ClassiqWiringError(
                f"Invalid in_wires type: {type(in_wires).__name__}"
            )

    def _connect_unnamed_in_quantum_registers(
        self,
        call: function_call.FunctionCall,
        quantum_registers: Collection[QuantumRegister],
    ) -> None:
        self._verify_legal_wires((i.wire for i in quantum_registers))

        call_inputs = call.function_params.get_io_names(
            function_params.IO.Input, call.is_inverse
        )

        if len(call_inputs) != len(quantum_registers):
            raise ClassiqWiringError(
                f'A call to "{call.name}" requires {len(call_inputs)} items, but {len(quantum_registers)} were given'
            )

        for input_name, quantum_register in zip(call_inputs, quantum_registers):
            quantum_register.consume()
            quantum_register.wire.connect_wire_end(end_call=call, input_name=input_name)

    def _connect_named_in_wires(
        self, call: function_call.FunctionCall, in_wires: Dict[str, QregOrWire]
    ) -> None:
        self._verify_legal_wires(
            wires=[
                in_wire.wire if isinstance(in_wire, QuantumRegister) else in_wire
                for in_wire in in_wires.values()
            ]
        )

        for input_name, in_wire in in_wires.items():
            if isinstance(in_wire, QuantumRegister):
                in_wire.consume()
                in_wire = in_wire.wire

            in_wire.connect_wire_end(end_call=call, input_name=input_name)

    def _connect_out_wires(
        self,
        call: function_call.FunctionCall,
        out_wires: SupportedIOUnion,
    ) -> Dict[str, Wire]:
        if isinstance(out_wires, dict):
            wire_dict = self._connect_named_out_wires(call=call, out_wires=out_wires)
        elif isinstance(out_wires, QuantumRegister):
            wire_dict = self._connect_unnamed_out_quantum_registers(
                call=call, quantum_registers=[out_wires]
            )
        elif isinstance(out_wires, collections.abc.Collection):
            if not all(isinstance(i, QuantumRegister) for i in out_wires):
                raise ClassiqWiringError(
                    "When supplying and iterable, all items must be instances of QReg"
                )
            wire_dict = self._connect_unnamed_out_quantum_registers(
                call=call, quantum_registers=out_wires  # type: ignore[arg-type]
            )
        else:
            raise ClassiqWiringError(
                f"Invalid out_wires type: {type(out_wires).__name__}"
            )

        self._update_generated_wires(wires=wire_dict.values())
        return wire_dict

    def _connect_named_out_wires(
        self,
        call: function_call.FunctionCall,
        out_wires: Dict[str, QregOrWire],
    ) -> Dict[str, Wire]:
        wire_dict: Dict[str, Wire] = {}
        output_names = call.function_params.get_io_names(
            function_params.IO.Output, call.is_inverse
        )

        for name in out_wires.keys():
            if name not in output_names:
                raise ClassiqWiringError(
                    f"output_name ({name}) does not belong to this function call"
                )

        for output_name in output_names:
            if output_name in out_wires:
                out_wire = out_wires[output_name]
                if isinstance(out_wire, QuantumRegister):
                    out_wire = out_wire.wire

                out_wire.connect_wire_start(start_call=call, output_name=output_name)
                wire_dict[output_name] = out_wire
            else:
                wire_dict[output_name] = self._output_wire_type(
                    start_call=call, output_name=output_name
                )

        return wire_dict

    def _connect_unnamed_out_quantum_registers(
        self,
        call: function_call.FunctionCall,
        quantum_registers: Collection[QuantumRegister],
    ) -> Dict[str, Wire]:
        wire_dict: Dict[str, Wire] = {}
        output_names = call.function_params.get_io_names(
            function_params.IO.Output, call.is_inverse
        )

        for quantum_register, output_name in zip(quantum_registers, output_names):
            quantum_register.wire.connect_wire_start(
                start_call=call, output_name=output_name
            )
            wire_dict[output_name] = quantum_register.wire

        return wire_dict

    def __getattr__(self, item):
        is_builtin_function_name = any(
            item == func.__name__
            for func in function_param_list.get_function_param_list()
        )

        if is_builtin_function_name:
            return functools.partial(self._function_call_handler, function=item)

        is_user_function_name = (
            self._function_library is not None
            and item in self._function_library.function_names
        )

        if is_user_function_name:
            return functools.partial(self.apply, function_name=item)

        raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{item}'")

    def __dir__(self):
        builtin_func_name = [
            func.__name__ for func in function_param_list.get_function_param_list()
        ]
        user_func_names = (
            list(self._function_library.function_names)
            if self._function_library is not None
            else list()
        )
        return list(super().__dir__()) + builtin_func_name + user_func_names

    def include_library(self, library: FunctionLibrary) -> None:
        """Includes a function library.

        Args:
            library (FunctionLibrary): The function library.
        """
        if self._function_library is not None:
            raise ClassiqValueError("Another function library is already included.")

        self._function_library = library

    @property
    @abc.abstractmethod
    def _logic_flow(self) -> List[function_call.FunctionCall]:
        pass

    @property
    @abc.abstractmethod
    def _output_wire_type(self) -> Type[Wire]:
        pass



import asyncio

from classiq_interface.chemistry import ground_state_problem, operator
from classiq_interface.chemistry.operator import PauliOperator

from classiq._internals import api_wrapper
from classiq.exceptions import ClassiqGenerationError


def generate_hamiltonian(
    gs_problem: ground_state_problem.GroundStateProblem,
) -> PauliOperator:
    return asyncio.run(generate_hamiltonian_async(gs_problem))


async def generate_hamiltonian_async(
    gs_problem: ground_state_problem.GroundStateProblem,
) -> PauliOperator:
    wrapper = api_wrapper.ApiWrapper()
    result = await wrapper.call_generate_hamiltonian_task(problem=gs_problem)

    if result.status != operator.OperatorStatus.SUCCESS:
        raise ClassiqGenerationError(f"Generate Hamiltonian failed: {result.details}")

    return result.details



"""Executor module, implementing facilities for executing quantum programs using Classiq platform."""
import asyncio
from typing import Union

import classiq_interface.executor.execution_preferences
from classiq_interface.executor import (
    execution_request,
    hamiltonian_minimization_problem,
    result as exc_result,
)
from classiq_interface.executor.result import (
    ExecutionDetails,
    FinanceSimulationResults,
    GroverSimulationResults,
)
from classiq_interface.executor.vqe_result import VQESolverResult
from classiq_interface.generator import result as generation_result

from classiq._internals import api_wrapper
from classiq._internals.type_validation import validate_type
from classiq.exceptions import ClassiqExecutionError


class Executor:
    """Executor is the entry point for executing quantum programs on multiple quantum hardware vendors."""

    def __init__(
        self,
        preferences: classiq_interface.executor.execution_preferences.ExecutionPreferences,
    ) -> None:
        """Init self.

        Args:
            preferences (): Execution preferences, such as number of shots.
        """
        self._preferences = preferences
        self._api_wrapper = api_wrapper.ApiWrapper()

    def execute_quantum_program(
        self, quantum_program: classiq_interface.executor.quantum_program.QuantumProgram
    ) -> ExecutionDetails:
        return asyncio.run(self.execute_quantum_program_async(quantum_program))

    async def execute_quantum_program_async(
        self, quantum_program: classiq_interface.executor.quantum_program.QuantumProgram
    ) -> ExecutionDetails:
        """Async version of `execute_quantum_program`"""
        request = execution_request.ExecutionRequest(
            execution_payload=execution_request.QuantumProgramExecution(
                **quantum_program.dict()
            ),
            preferences=self._preferences,
        )
        execution_result = await self._api_wrapper.call_execute_task(request=request)

        if execution_result.status != exc_result.ExecutionStatus.SUCCESS:
            raise ClassiqExecutionError(f"Execution failed: {execution_result.details}")
        return validate_type(
            obj=execution_result.details,
            expected_type=ExecutionDetails,
            operation="Execution",
            exception_type=ClassiqExecutionError,
        )

    def execute_generated_circuit(
        self, generation_result: generation_result.GeneratedCircuit
    ) -> Union[FinanceSimulationResults, GroverSimulationResults]:
        return asyncio.run(self.execute_generated_circuit_async(generation_result))

    async def execute_generated_circuit_async(
        self, generation_result: generation_result.GeneratedCircuit
    ) -> Union[FinanceSimulationResults, GroverSimulationResults]:
        """Async version of `execute_generated_circuit`"""
        if generation_result.metadata is None:
            raise ClassiqExecutionError(
                "The execute_generated_circuit is to execute generated circuits as oracles, but "
                "the generated circuit's metadata is empty. To execute a circuit as-is, please"
                "use execute_quantum_program."
            )
        request = execution_request.ExecutionRequest(
            execution_payload=execution_request.GenerationMetadataExecution(
                **generation_result.metadata.dict()
            ),
            preferences=self._preferences,
        )
        execution_result = await self._api_wrapper.call_execute_task(request=request)

        if execution_result.status != exc_result.ExecutionStatus.SUCCESS:
            raise ClassiqExecutionError(f"Execution failed: {execution_result.details}")
        return validate_type(
            obj=execution_result.details,
            expected_type=(FinanceSimulationResults, GroverSimulationResults),
            operation="Execution",
            exception_type=ClassiqExecutionError,
        )

    def execute_hamiltonian_minimization(
        self,
        hamiltonian_minimization_problem: hamiltonian_minimization_problem.HamiltonianMinimizationProblem,
    ) -> VQESolverResult:
        return asyncio.run(
            self.execute_hamiltonian_minimization_async(
                hamiltonian_minimization_problem
            )
        )

    async def execute_hamiltonian_minimization_async(
        self,
        hamiltonian_minimization_problem: hamiltonian_minimization_problem.HamiltonianMinimizationProblem,
    ) -> VQESolverResult:
        """Async version of `execute_hamiltonian_minimization`"""
        request = execution_request.ExecutionRequest(
            execution_payload=execution_request.HamiltonianMinimizationProblemExecution(
                **hamiltonian_minimization_problem.dict()
            ),
            preferences=self._preferences,
        )
        execution_result = await self._api_wrapper.call_execute_task(request=request)

        if execution_result.status != exc_result.ExecutionStatus.SUCCESS:
            raise ClassiqExecutionError(f"Execution failed: {execution_result.details}")
        return validate_type(
            obj=execution_result.details,
            expected_type=VQESolverResult,
            operation="Execution",
            exception_type=ClassiqExecutionError,
        )



import asyncio

from classiq._internals.client import client


def authenticate(overwrite: bool = False):
    asyncio.run(authenticate_async(overwrite))


async def authenticate_async(overwrite: bool = False):
    """Async version of `register_device`"""
    await client().authenticate(overwrite)



import urllib.parse
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from httpx import AsyncClient, Response, codes

from classiq_interface.server import authentication

from classiq.exceptions import ClassiqAuthenticationError


@dataclass
class Tokens:
    access_token: str
    refresh_token: Optional[str]


class Auth0:
    _SDK_CLIENT_ID = "f6721qMOVoDAOVkzrv8YaWassRKSFX6Y"
    _CONTENT_TYPE = "application/x-www-form-urlencoded"
    _HEADERS = {"content-type": _CONTENT_TYPE}
    _BASE_URL = f"https://{authentication.AUTH0_DOMAIN}"

    @classmethod
    async def _make_request(
        cls,
        url: str,
        payload: Dict[str, str],
        allow_error: Union[bool, int] = False,
    ) -> Dict[str, Any]:
        encoded_payload = urllib.parse.urlencode(payload)
        client: AsyncClient
        async with AsyncClient(base_url=cls._BASE_URL, headers=cls._HEADERS) as client:
            response: Response = await client.post(url=url, content=encoded_payload)
            code = response.status_code
            error_code_allowed = allow_error is True or allow_error == code
            data = response.json()

        if code == codes.OK or error_code_allowed:
            return data

        raise ClassiqAuthenticationError(
            f"Request to Auth0 failed with error code {code}: {data.get('error')}"
        )

    @classmethod
    async def get_device_data(cls, get_refresh_token: bool = True) -> Dict[str, Any]:
        payload = {
            "client_id": cls._SDK_CLIENT_ID,
            "audience": authentication.API_AUDIENCE,
        }
        if get_refresh_token:
            payload["scope"] = "offline_access"

        return await cls._make_request(
            url="/oauth/device/code",
            payload=payload,
        )

    @classmethod
    async def poll_tokens(cls, device_code: str) -> Dict[str, Any]:
        payload = {
            "client_id": cls._SDK_CLIENT_ID,
            "device_code": device_code,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }

        return await cls._make_request(
            url="/oauth/token",
            payload=payload,
            allow_error=codes.FORBIDDEN,
        )

    @classmethod
    async def refresh_access_token(cls, refresh_token: str) -> Tokens:
        # TODO handle failure
        payload = {
            "client_id": cls._SDK_CLIENT_ID,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }
        data = await cls._make_request(
            url="/oauth/token",
            payload=payload,
        )

        return Tokens(
            access_token=data["access_token"],
            refresh_token=data.get("refresh_token", None),
        )



from typing import Optional

from classiq_interface.generator import function_call

from classiq.exceptions import ClassiqWiringError


class Wire:
    _start_call: Optional[function_call.FunctionCall] = None
    _start_name: Optional[str] = None
    _end_call: Optional[function_call.FunctionCall] = None

    def __init__(
        self,
        start_call: Optional[function_call.FunctionCall] = None,
        output_name: str = "",
    ) -> None:
        self.connect_wire_start(start_call=start_call, output_name=output_name)  # type: ignore[arg-type]

    def _initialize_wire_name(self, wire_name: Optional[str] = None) -> None:
        if wire_name is not None:
            self._wire_name = wire_name
        elif self.is_started:
            self._wire_name = f"{self._start_call.name}_{self._start_name}"  # type: ignore[union-attr]
        else:
            self._wire_name = ""

    def connect_wire_start(
        self, start_call: function_call.FunctionCall, output_name: str
    ) -> None:
        if self.is_started:
            raise ClassiqWiringError(
                "Cannot connect wire-start to an already started wire."
            )

        self._start_call: function_call.FunctionCall = start_call
        self._start_name: str = output_name

        self._initialize_wire_name()

    def connect_wire_end(
        self, end_call: function_call.FunctionCall, input_name: str
    ) -> None:
        if self.is_connected:
            raise ClassiqWiringError("Wire already connected")

        self._end_call: function_call.FunctionCall = end_call
        self._end_name: str = input_name

        end_call.inputs[input_name] = self._wire_name

        if self._start_call is not None:
            self.set_as_output(self._wire_name)

    def set_as_output(self, output_name: str) -> None:
        if not self.is_started:
            raise ClassiqWiringError("Wire initialized incorrectly")
        if self.is_connected:
            raise ClassiqWiringError("Wire already connected")
        if self._start_name is None:
            raise ClassiqWiringError("Wire start name is uninitialized")

        self._start_call.outputs[self._start_name] = output_name  # type: ignore[union-attr]

    @property
    def is_connected(self) -> bool:
        return self.is_started and self.is_ended

    @property
    def is_started(self) -> bool:
        return self._start_call is not None

    @property
    def is_ended(self) -> bool:
        return (self._end_call is not None) and (
            self.is_started and self._start_name in self._start_call.outputs  # type: ignore[union-attr]
        )



import logging

_PACKAGE_NAME = __name__.split(".", maxsplit=1)[0]


def add_stderr_logger(level: int = logging.DEBUG) -> logging.StreamHandler:
    """
    Helper for quickly adding a StreamHandler to the logger. Useful for
    debugging.
    Returns the handler after adding it.
    """
    logger = logging.getLogger(_PACKAGE_NAME)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")
    )
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.debug("Added a stderr logging handler to logger: %s", logger.name)
    return handler



"""Classiq SDK."""

from classiq._internals import logger  # noqa: F401
from classiq._internals.async_utils import (
    enable_jupyter_notebook,
    is_notebook as _is_notebook,
)
from classiq._internals.authentication.authentication import authenticate  # noqa: F401
from classiq._internals.client import configure  # noqa: F401
from classiq._internals.config import Configuration  # noqa: F401
from classiq._version import VERSION as _VERSION
from classiq.analyzer import Analyzer  # noqa: F401
from classiq.applications import *  # noqa: F401, F403
from classiq.executor import Executor  # noqa: F401
from classiq.model_designer import *  # noqa: F401, F403
from classiq.quantum_functions import *  # noqa: F401, F403
from classiq.quantum_register import *  # noqa: F401, F403

__version__ = _VERSION


if _is_notebook():
    enable_jupyter_notebook()



import functools
import logging
import os
import platform
import ssl
import sys
import urllib.parse
from typing import Any, Dict, Optional, Tuple, Union

import httpx
import pydantic
import websockets.client
from semver import VersionInfo

from classiq_interface import __version__ as classiq_interface_version
from classiq_interface.server import authentication
from classiq_interface.server.routes import ROUTE_PREFIX

from classiq._internals import config
from classiq._internals.authentication import token_manager
from classiq._version import VERSION as CLASSIQ_VERSION
from classiq.exceptions import ClassiqAPIError, ClassiqExpiredTokenError

_PROCESSED_VERSION = Union[VersionInfo, str]
_PROCESSED_VERSIONS = Tuple[_PROCESSED_VERSION, _PROCESSED_VERSION]


_VERSION_UPDATE_SUGGESTION = (
    'Run "pip install -U <PACKAGE>==<REQUIRED VERSION>" to resolve the conflict.'
)
_FRONTEND_VARIANT: str = "classiq-sdk"
_INTERFACE_VARIANT: str = "classiq-interface-sdk"
_USERAGENT_SEPARATOR: str = " "

_logger = logging.getLogger(__name__)


@functools.lru_cache()
def _get_python_execution_environment() -> str:
    # Try spyder
    if any("SPYDER" in name for name in os.environ):
        return "Spyder"

    # try ipython and its variants
    try:
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined]
        if shell == "ZMQInteractiveShell":
            return "Jupyter"  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return "IPython"  # Terminal running IPython
        else:
            return "IPython-other"  # Other type (?)
    except NameError:
        pass  # Probably standard Python interpreter

    # Try VSCode
    if "debugpy" in sys.modules:
        return "VSCode"

    # Try pycharm
    if "PYCHARM_HOSTED" in os.environ:
        return "PyCharm"

    return "Python"


@functools.lru_cache()
def _get_user_agent_header() -> Dict[str, str]:
    python_version = (
        f"python({_get_python_execution_environment()})/{platform.python_version()}"
    )
    os_platform = f"{os.name}/{platform.platform()}"
    frontend_version = f"{_FRONTEND_VARIANT}/{CLASSIQ_VERSION}"
    interface_version = f"{_INTERFACE_VARIANT}/{classiq_interface_version}"
    return {
        "User-Agent": _USERAGENT_SEPARATOR.join(
            (python_version, os_platform, frontend_version, interface_version)
        )
    }


class HostVersions(pydantic.BaseModel):
    classiq_interface: pydantic.StrictStr = pydantic.Field()


class Client:
    _UNKNOWN_VERSION = "0.0.0"

    def __init__(self, conf: config.Configuration):
        self._config = conf
        self._token_manager = token_manager.TokenManager()
        self._ssl_context = ssl.create_default_context()
        self._HTTP_TIMEOUT_SECONDS = (
            3600  # Needs to be synced with load-balancer timeout
        )

    @staticmethod
    def _handle_response(response: httpx.Response) -> Union[Dict, str]:
        if response.is_error:
            expired = (
                response.status_code == httpx.codes.UNAUTHORIZED
                and response.json()["detail"] == authentication.EXPIRED_TOKEN_ERROR
            )

            if expired:
                raise ClassiqExpiredTokenError("Expired token.")

            try:
                message = (
                    f"Call to API failed with code {response.status_code}: "
                    f"{response.json()['detail']}"
                )
            except BaseException:
                message = "Call to API failed"
            raise ClassiqAPIError(message)

        return response.json()

    def _make_client_args(self) -> Dict[str, Any]:
        return {
            "base_url": self._config.host,
            "timeout": self._HTTP_TIMEOUT_SECONDS,
            "headers": {
                **self._get_authorization_header(),
                **_get_user_agent_header(),
            },
        }

    async def call_api(
        self, http_method: str, url: str, body: Optional[Dict] = None
    ) -> Union[Dict, str]:
        async with httpx.AsyncClient(**self._make_client_args()) as async_client:
            response = await async_client.request(
                method=http_method, url=url, json=body
            )
            return self._handle_response(response)

    def sync_call_api(
        self, http_method: str, url: str, body: Optional[Dict] = None
    ) -> Union[Dict, str]:
        with httpx.Client(**self._make_client_args()) as sync_client:
            response = sync_client.request(method=http_method, url=url, json=body)
            return self._handle_response(response)

    def _get_authorization_header(self) -> Dict[str, str]:
        access_token = self._token_manager.get_access_token()
        if access_token is None:
            return dict()
        return {"Authorization": f"Bearer {access_token}"}

    def _get_authorization_query_string(self) -> str:
        access_token = self._token_manager.get_access_token()
        if access_token is None:
            return ""
        return urllib.parse.urlencode({"token": access_token})

    async def update_expired_access_token(self) -> None:
        await self._token_manager.update_expired_access_token()

    def establish_websocket_connection(
        self, path: str, **kwargs
    ) -> websockets.client.connect:
        _MAX_PAYLOAD_SIZE = 2**23  # = 8MiB ~= 8MB
        max_size = kwargs.get("max_size", _MAX_PAYLOAD_SIZE)
        ping_interval = kwargs.get("ping_interval", None)
        url = urllib.parse.urlsplit(self._config.ws_uri)
        url = url._replace(path=path, query=self._get_authorization_query_string())

        return websockets.client.connect(
            uri=url.geturl(),
            ssl=self._ssl_context if self._config.ws_uri.scheme == "wss" else None,
            max_size=max_size,
            ping_interval=ping_interval,
            extra_headers=_get_user_agent_header(),
        )

    def get_backend_uri(self) -> str:
        return self._config.host

    def _get_host_version(self) -> str:
        host = HostVersions.parse_obj(
            self.sync_call_api("get", f"{ROUTE_PREFIX}/versions")
        )
        return host.classiq_interface

    @classmethod
    def _check_matching_versions(
        cls, lhs_version: str, rhs_version: str, normalize: bool = True
    ) -> bool:
        processed_versions: _PROCESSED_VERSIONS
        if normalize:
            # VersionInfo comparison is compatible with strings but it excludes any build info
            processed_versions = VersionInfo.parse(lhs_version), VersionInfo.parse(
                rhs_version
            )
        else:
            processed_versions = lhs_version, rhs_version
        if cls._UNKNOWN_VERSION in processed_versions:
            # In case one of those versions is unknown, they are considered equal
            _logger.debug(
                "Either {} or {} is an unknown version. Assuming both versions are equal.",
                lhs_version,
                rhs_version,
            )
            return True
        return processed_versions[0] == processed_versions[1]

    def check_host(self) -> None:
        # This function is NOT async (despite the fact that it can be) because it's called from a non-async context.
        # If this happens when we already run in an event loop (e.g. inside a call to asyncio.run), we can't go in to
        # an async context again.
        # Since this function should be called ONCE in each session, we can handle the "cost" of blocking the
        # event loop.
        if not self._check_matching_versions(
            classiq_interface_version, CLASSIQ_VERSION, normalize=False
        ):
            # When raising an exception, use the original strings
            raise ClassiqAPIError(
                f"Classiq API version mismatch: 'classiq' version is {CLASSIQ_VERSION}, "
                f"'classiq-interface' version is {classiq_interface_version}. {_VERSION_UPDATE_SUGGESTION}"
            )

        try:
            raw_host_version = self._get_host_version()
        except httpx.ConnectError:
            _logger.warning(
                "Version check failed - host unavailable.",
            )
        else:
            if not self._check_matching_versions(
                raw_host_version, classiq_interface_version
            ):
                raise ClassiqAPIError(
                    f"Classiq API version mismatch: 'classiq-interface' version is "
                    f"{classiq_interface_version}, backend version is {raw_host_version}. {_VERSION_UPDATE_SUGGESTION}"
                )

    async def authenticate(self, overwrite: bool) -> None:
        await self._token_manager.manual_authentication(overwrite=overwrite)


DEFAULT_CLIENT: Optional[Client] = None


def client() -> Client:
    global DEFAULT_CLIENT
    if DEFAULT_CLIENT is None:
        # This call initializes DEFAULT_CLIENT
        configure(config.init())
    assert DEFAULT_CLIENT is not None
    return DEFAULT_CLIENT


def configure(conf: config.Configuration) -> None:
    global DEFAULT_CLIENT
    assert DEFAULT_CLIENT is None, "Can not configure client after first usage."

    DEFAULT_CLIENT = Client(conf=conf)
    if conf.should_check_host:
        DEFAULT_CLIENT.check_host()



import argparse
import logging
import os
import threading
import warnings
from typing import Optional

from classiq._internals.authentication import password_manager as pm
from classiq._internals.authentication.auth0 import Auth0
from classiq._internals.authentication.device import DeviceRegistrar, Tokens
from classiq.exceptions import ClassiqAuthenticationError

_logger = logging.getLogger(__name__)


class TokenManager:
    def __init__(self, password_manager: Optional[pm.PasswordManager] = None):
        if password_manager is None:
            password_manager = self._make_password_manager()
        # We use threading.Lock instead of asyncio.Lock because the latter is coupled
        # to a specific event loop, which is undesirable
        self._lock = threading.Lock()
        self._password_manager: pm.PasswordManager = password_manager
        self._access_token: Optional[str] = self._password_manager.access_token
        self._refresh_token: Optional[str] = self._password_manager.refresh_token
        if self._access_token is None and self._refresh_token is not None:
            _logger.info(
                "Inconsistent state, access token is absent and refresh token is present."
            )

    @classmethod
    def _make_password_manager(cls) -> pm.PasswordManager:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--skip-authentication", action="store_true", required=False
        )
        args, _ = parser.parse_known_args()

        use_dummy = args.skip_authentication or os.getenv("PYTEST_RUNNING") == "true"

        if use_dummy:
            return pm.DummyPasswordManager()
        else:
            return pm.PasswordManager()

    def get_access_token(self) -> Optional[str]:
        return self._access_token

    async def _refresh(self) -> None:
        assert self._refresh_token is not None
        tokens = await Auth0.refresh_access_token(self._refresh_token)
        self._save_tokens(tokens)

    async def update_expired_access_token(self) -> None:
        if self._refresh_token is not None:
            await self._refresh()
            return
        self._clear_tokens()
        raise ClassiqAuthenticationError(
            "Please call `classiq.authenticate()` to manually log in."
        )

    async def manual_authentication(self, overwrite: bool) -> None:
        if self._refresh_token is None:
            await self._authentication_helper()
            return
        # Note, this function warns using warnings because https://stackoverflow.com/q/9595009
        if overwrite:
            warnings.warn(
                "Overwriting an existing refresh token should only be done if "
                "it is compromised. Make sure this operation is necessary, "
                "and if not, remove the call to device registration."
            )
            await self._authentication_helper()
        else:
            warnings.warn(
                "Device is already registered.\nGenerating a new refresh token should only "
                "be done if the current refresh token is compromised.\nTo do so, set the "
                "overwrite parameter to true"
            )
            if self._access_token is None:
                await self._refresh()

    def _save_tokens(
        self, tokens: Tokens, force_override_refresh_token: bool = False
    ) -> None:
        with self._lock:
            self._access_token = tokens.access_token
            self._password_manager.access_token = tokens.access_token
            if tokens.refresh_token is not None or force_override_refresh_token:
                self._refresh_token = tokens.refresh_token
                self._password_manager.refresh_token = tokens.refresh_token

    def _clear_tokens(self) -> None:
        with self._lock:
            self._access_token = None
            self._password_manager.access_token = None
            self._refresh_token = None
            self._password_manager.refresh_token = None

    async def _authentication_helper(self) -> None:
        # TODO: consider using refresh token rotation
        #  (https://auth0.com/docs/tokens/refresh-tokens/refresh-token-rotation)
        tokens = await DeviceRegistrar.register(
            get_refresh_token=pm.is_password_manager_available()
        )
        self._save_tokens(tokens, force_override_refresh_token=True)



def enable_jupyter_notebook() -> None:
    import nest_asyncio  # type: ignore[import]

    nest_asyncio.apply()


# =======================================================================
# According to stackoverflow.com's license
# taken from:
#   https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
# from the user:
#   https://stackoverflow.com/users/2132753/gustavo-bezerra
def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined]
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


# =======================================================================



from typing import List, Optional, _GenericAlias  # type: ignore[attr-defined]

from classiq_interface.generator.arith.arithmetic import RegisterUserInput
from classiq_interface.generator.synthesis_metrics import WireRole as Role

from classiq.exceptions import ClassiqQRegError
from classiq.model_designer.wire import Wire


# This class is used for QReg, to support type-hint initialization
#   Due to the `size` property of QReg
class _GenericAliasWithSize(_GenericAlias, _root=True):  # type: ignore[call-arg]
    def __call__(self, *args, **kwargs):
        if self.size is not None:
            return super().__call__(*args, size=self.size, **kwargs)
        else:
            return super().__call__(*args, **kwargs)

    @property
    def role(self) -> Optional[Role]:
        return getattr(self.__origin__, "role", None)

    @property
    def size(self) -> Optional[int]:
        if len(self.__args__) == 1 and type(self.__args__[0]) is int:
            return self.__args__[0]
        return None


# This is a private class, whose purpose is to support slicing QuantumRegisters,
#   And to keep some link between the qubits.
class _Qubit:
    _is_available = True

    def __bool__(self) -> bool:
        return self._is_available

    def __repr__(self) -> str:
        status = "available" if self._is_available else "consumed"
        return f"{self.__class__.__name__[1:]}({status})"

    @property
    def is_available(self) -> bool:
        return self._is_available

    @property
    def is_consumed(self) -> bool:
        return not self.is_available

    def consume(self) -> None:
        if self.is_consumed:
            raise ClassiqQRegError("Qubit is already consumed")
        self._is_available = False


class QReg:
    """A Quantum Register - A logical collection of several qubits.

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section. Alternatively, attributes may be documented
    inline with the attribute's declaration (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.

    Attributes:
        size (int): The amount of qubits.

    """

    is_signed: Optional[bool] = None
    fraction_places: Optional[int] = None

    # Object initialization
    def __init__(self, size: int) -> None:
        self.size: int = size

        self._wire: Wire = Wire()

        self._available_qubits: List[_Qubit] = [_Qubit() for _ in range(self.size)]

    def __class_getitem__(cls, params):
        # Supporting python 3.7+, thus returning `typing._GenericAlias` instead of `types.GenericAlias`
        if type(params) is int:
            return _GenericAliasWithSize(cls, params, inst=True, special=False)

        raise ClassiqQRegError(f"Invalid size: {params} ; int required")

    # Exported functions
    def __len__(self) -> int:
        return self.size

    # Exported functions - Availability
    @property
    def is_available(self) -> bool:
        """bool: Checks whether this QReg is available."""
        return all(self._available_qubits)

    @property
    def is_consumed(self) -> bool:
        """bool: Checks whether this QReg is consumed."""
        return not self.is_available

    def get_available_indexes(self) -> List[int]:
        """List[int]: Returns the indexes of the qubits that are available."""
        return [index for index, value in enumerate(self._available_qubits) if value]

    def consume(self) -> None:
        # This function is for internal usage.
        if self.is_consumed:
            raise ClassiqQRegError("Cannot consume a consumed QReg")

        for qubit in self._available_qubits:
            qubit.consume()

    # Wiring
    @property
    def wire(self) -> Wire:
        """Wire: Returns the wire associated with this QReg."""
        return self._wire


# QReg with arithmetic properties
class QSFixed(QReg):
    is_signed: bool = True

    def __init__(self, size: int, fraction_places: int) -> None:
        self.fraction_places: int = fraction_places
        super().__init__(size=size)

    def to_register_user_input(self) -> RegisterUserInput:
        return RegisterUserInput(
            size=self.size,
            is_signed=self.is_signed,
            fraction_places=self.fraction_places,
        )


QFixed = QSFixed


class QUFixed(QFixed):
    is_signed: bool = False


class QSInt(QFixed):
    def __init__(self, size: int):
        super().__init__(size=size, fraction_places=0)


QInt = QSInt


class QUInt(QInt):
    is_signed: bool = False


# QReg with synthesis properties
class ZeroQReg(QReg):
    role: Role = Role.ZERO


class AuxQReg(ZeroQReg):
    role: Role = Role.AUXILIARY



"""Analyzer module, implementing facilities for analyzing circuits using Classiq platform."""
import asyncio
import webbrowser

from classiq_interface.analyzer import analysis_params, result as analysis_result
from classiq_interface.generator import result as generator_result
from classiq_interface.server import routes

from classiq._internals import api_wrapper, client
from classiq._internals.type_validation import validate_type
from classiq.exceptions import ClassiqAnalyzerError


class Analyzer:
    """Analyzer is the wrapper object for all analysis capabilities."""

    def __init__(self, circuit: generator_result.GeneratedCircuit):
        """Init self.

        Args:
            circuit (): The circuit to be analyzed.
        """
        if circuit.qasm is None:
            raise ValueError("Analysis requires a circuit with valid QASM code")
        self._params = analysis_params.AnalysisParams(qasm=circuit.qasm)

    def analyze(self) -> analysis_result.Analysis:
        """Runs the circuit analysis.

        Returns:
            The analysis result.
        """
        return asyncio.run(self.analyze_async())

    async def analyze_async(self) -> analysis_result.Analysis:
        """Async version of `analyze`
        Runs the circuit analysis.

        Returns:
            The analysis result.
        """
        wrapper = api_wrapper.ApiWrapper()
        result = await wrapper.call_analysis_task(params=self._params)

        if result.status != analysis_result.AnalysisStatus.SUCCESS:
            raise ClassiqAnalyzerError(f"Analysis failed: {result.details}")
        details = validate_type(
            obj=result.details,
            expected_type=analysis_result.Analysis,
            operation="Analysis",
            exception_type=ClassiqAnalyzerError,
        )

        dashboard_path = routes.ANALYZER_DASHBOARD
        self.run_external_app(path=dashboard_path)
        return details

    @staticmethod
    def run_external_app(path: str) -> None:
        backend_uri = client.client().get_backend_uri()
        webbrowser.open_new_tab(f"{backend_uri}{path}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['classiq',
 'classiq._internals',
 'classiq._internals.authentication',
 'classiq.applications',
 'classiq.builtin_functions',
 'classiq.model_designer',
 'classiq.quantum_functions']

package_data = \
{'': ['*'], 'classiq': ['examples/*']}

install_requires = \
['ConfigArgParse>=1.5.3,<2.0.0',
 'Pyomo>=6.0,<6.1',
 'httpx>=0.22.0,<1',
 'keyring>=23.5.0,<24.0.0',
 'nest-asyncio>=1.5.4,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'semver>=2.13.0,<3.0.0',
 'websockets>=10.1,<11.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.11.1,<5.0.0'],
 'all': ['classiq-interface>=0.7.1,<0.8.0']}

setup_kwargs = {
    'name': 'classiq',
    'version': '0.7.1',
    'description': "Classiq's Python SDK for quantum computing",
    'long_description': '<p align="center">\n  <a href="https://www.classiq.io"><img src="https://uploads-ssl.webflow.com/60000db7a5f449af5e4590ac/6122b22eea7a9583a5c0d560_classiq_RGB_Green_with_margin.png\n" alt="Classiq"></a>\n</p>\n<p align="center">\n    <em>The Classiq Quantum Algorithm Design platform helps teams build sophisticated quantum circuits that could not be designed otherwise</em>\n</p>\n\nWe do this by synthesizing high-level functional models into optimized quantum\ncircuits, taking into account the constraints that are important to the designer.\nFurthermore, we are able to generate circuits for practically any universal gate-based\nquantum computer and are compatible with most quantum cloud providers.\n\nSee the\n[documentation](https://classiquantum.com/reference/getting-started/python-sdk.html)\nfor information about this extension.\n\n## License\n\nSee [license](https://classiq.io/license).\n',
    'author': 'Classiq Technologies',
    'author_email': 'support@classiq.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://classiq.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Function library module, implementing facilities for adding user defined functions to the Classiq platform."""

from typing import Dict, Tuple, Union

from classiq_interface.generator.functions import (
    DEFAULT_FUNCTION_LIBRARY_NAME,
    FunctionData,
    FunctionLibraryData,
    FunctionType,
)
from classiq_interface.generator.user_defined_function_params import CustomFunction

from classiq.exceptions import ClassiqValueError
from classiq.quantum_functions.quantum_function import QuantumFunction

QASM_INTRO = 'OPENQASM 2.0;\ninclude "qelib1.inc";\n'


class FunctionLibrary:
    """Facility to manage functions."""

    def __init__(self, *functions, name: str = DEFAULT_FUNCTION_LIBRARY_NAME):
        """
        Args:
            name (:obj:`str`, optional): The name of the function library.
            *functions (:obj:`FunctionData`, optional): A list of functions to initialize the object.
        """
        self._data = FunctionLibraryData(name=name)
        self._params: Dict[str, CustomFunction] = dict()

        for f in functions:
            self.add_function(f)

    def get_function(self, function_name: str) -> CustomFunction:
        return self._params[function_name]

    def add_function(
        self,
        function_data: Union[FunctionData, QuantumFunction],
        override_existing_functions: bool = False,
    ) -> CustomFunction:
        """Adds a function to the function library.

        Args:
            function_data (FunctionData): The function data object.
            override_existing_functions (:obj:`bool`, optional): Defaults to False.

        Returns:
            The custom function parameters.
        """
        if isinstance(function_data, QuantumFunction):
            function_data = function_data.function_data

        if not isinstance(function_data, FunctionData):
            raise ClassiqValueError(
                f"FunctionData object expected, got {function_data.__class__.__name__}"
            )

        function_name = function_data.name
        if (
            not override_existing_functions
            and function_name in self._data.function_dict
        ):
            raise ClassiqValueError("Cannot override existing functions.")

        if function_data.function_type == FunctionType.CompositeFunction:
            for call in function_data.logic_flow:
                if not isinstance(call.function_params, CustomFunction):
                    continue
                FunctionLibraryData.validate_function_in_library(
                    library=self._data,
                    function_params=call.function_params,
                    error_handler=ClassiqValueError,
                )

        self._data.function_dict[function_name] = function_data
        self._params[function_name] = self._to_params(function_data)
        return self.get_function(function_name=function_name)

    def remove_function(self, function_name: str) -> FunctionData:
        """Removes a function from the function library.

        Args:
            function_name (str): The name of the function.

        Returns:
            The removed function data.
        """
        self._params.pop(function_name)
        return self._data.function_dict.pop(function_name)

    @property
    def name(self) -> str:
        """The library name."""
        return self._data.name

    @property
    def function_names(self) -> Tuple[str, ...]:
        """Get a tuple of the names of the functions in the library.

        Returns:
            The names of the functions in the library.
        """
        return tuple(self._data.function_dict.keys())

    @property
    def data(self) -> FunctionLibraryData:
        return self._data

    @staticmethod
    def _to_params(data: FunctionData) -> CustomFunction:
        params = CustomFunction(name=data.name)
        params.generate_io_names(
            input_set=data.input_set,
            output_set=data.output_set,
        )
        return params



from typing import Callable

from classiq.quantum_functions.quantum_function import QuantumFunction


def quantum_function(func: Callable) -> QuantumFunction:
    qf = QuantumFunction()
    qf.add_implementation(func)
    return qf



import logging

_logger = logging.getLogger(__name__)


class ClassiqError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        _logger.error("%s\n", message)


class ClassiqExecutionError(ClassiqError):
    pass


class ClassiqGenerationError(ClassiqError):
    pass


class ClassiqAnalyzerError(ClassiqError):
    pass


class ClassiqAPIError(ClassiqError):
    pass


class ClassiqVersionError(ClassiqError):
    pass


class ClassiqValueError(ClassiqError, ValueError):
    pass


class ClassiqWiringError(ClassiqValueError):
    pass


class ClassiqQRegError(ClassiqValueError):
    pass


class ClassiqQFuncError(ClassiqValueError):
    pass


class ClassiqAuthenticationError(ClassiqError):
    pass


class ClassiqExpiredTokenError(ClassiqAuthenticationError):
    pass


class ClassiqFileNotFoundError(FileNotFoundError):
    pass



from typing import Optional, Dict

from classiq_interface.generator.function_param_list import *

from classiq_interface.generator.standard_gates.controlled_standard_gates import *

from classiq_interface.generator.standard_gates.standard_angled_gates import *

from classiq_interface.generator.standard_gates.standard_gates import *

import abc
from classiq.exceptions import ClassiqValueError as ClassiqValueError, ClassiqWiringError as ClassiqWiringError
from classiq.model_designer.wire import Wire as Wire
from classiq.quantum_functions.function_library import FunctionLibrary as FunctionLibrary
from classiq.quantum_register import QReg as QuantumRegister
from typing import Collection, Dict, Iterable, Optional, Union

QregOrWire = Union[QuantumRegister, Wire]
WireOrWires = Union[Wire, Iterable[Wire]]
SupportedIOUnion = Union[Dict[str, QregOrWire], Collection[QuantumRegister], QuantumRegister]

class FunctionHandler(abc.ABC, metaclass=abc.ABCMeta):
    def __init__(self) -> None: ...
    def apply(self, function_name: str, in_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = ..., out_wires: Optional[Union[Dict[str, QregOrWire], QuantumRegister]] = ...) -> Dict[str, Wire]: ...
    def __getattr__(self, item): ...
    def __dir__(self): ...
    def include_library(self, library: FunctionLibrary) -> None: ...
    def RYGate(self, params: RYGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Negation(self, params: Negation, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def GreaterThan(self, params: GreaterThan, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Subtractor(self, params: Subtractor, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LogicalAnd(self, params: LogicalAnd, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def ZGate(self, params: ZGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def TGate(self, params: TGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def UnitaryGate(self, params: UnitaryGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SwapGate(self, params: SwapGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CHGate(self, params: CHGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RZZGate(self, params: RZZGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def FinanceModels(self, params: FinanceModels, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def BitwiseXor(self, params: BitwiseXor, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def TdgGate(self, params: TdgGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SXdgGate(self, params: SXdgGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def StatePreparation(self, params: StatePreparation, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def C3XGate(self, params: C3XGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CXGate(self, params: CXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def ArithmeticOracle(self, params: ArithmeticOracle, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RXGate(self, params: RXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RYYGate(self, params: RYYGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CSXGate(self, params: CSXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SXGate(self, params: SXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def YGate(self, params: YGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Min(self, params: Min, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def HadamardAmpLoad(self, params: HadamardAmpLoad, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def AmplitudeEstimation(self, params: AmplitudeEstimation, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CRXGate(self, params: CRXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LShift(self, params: LShift, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LessThan(self, params: LessThan, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def C4XGate(self, params: C4XGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def BitwiseOr(self, params: BitwiseOr, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LinearPauliRotations(self, params: LinearPauliRotations, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def HypercubeEntangler(self, params: HypercubeEntangler, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def PhaseGate(self, params: PhaseGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def BitwiseInvert(self, params: BitwiseInvert, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def StatePropagator(self, params: StatePropagator, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def WeightedAdder(self, params: WeightedAdder, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RXXGate(self, params: RXXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Mcx(self, params: Mcx, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RShift(self, params: RShift, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SGate(self, params: SGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def BitwiseAnd(self, params: BitwiseAnd, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def GroverOperator(self, params: GroverOperator, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def HartreeFock(self, params: HartreeFock, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SdgGate(self, params: SdgGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CRYGate(self, params: CRYGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def XGate(self, params: XGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RGate(self, params: RGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CRZGate(self, params: CRZGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def GreaterEqual(self, params: GreaterEqual, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def MCPhaseGate(self, params: MCPhaseGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Exponentiation(self, params: Exponentiation, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LinearGCI(self, params: LinearGCI, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Arithmetic(self, params: Arithmetic, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LessEqual(self, params: LessEqual, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Finance(self, params: Finance, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def IGate(self, params: IGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def iSwapGate(self, params: iSwapGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CCXGate(self, params: CCXGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def RZGate(self, params: RZGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def TwoDimensionalEntangler(self, params: TwoDimensionalEntangler, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CDFComparator(self, params: CDFComparator, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Equal(self, params: Equal, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def FinancePayoff(self, params: FinancePayoff, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Multiplier(self, params: Multiplier, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def QFT(self, params: QFT, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def LogicalOr(self, params: LogicalOr, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def UCC(self, params: UCC, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def HGate(self, params: HGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def NotEqual(self, params: NotEqual, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CZGate(self, params: CZGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Max(self, params: Max, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CYGate(self, params: CYGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CustomFunction(self, params: CustomFunction, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def Adder(self, params: Adder, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def SparseAmpLoad(self, params: SparseAmpLoad, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def GridEntangler(self, params: GridEntangler, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CPhaseGate(self, params: CPhaseGate, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def HardwareEfficientAnsatz(self, params: HardwareEfficientAnsatz, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...
    def CyclicShift(self, params: CyclicShift, in_wires: Optional[Dict[str, Wire]] = None) -> Dict[str, Wire]: ...



import asyncio
import time
import webbrowser
from datetime import timedelta
from typing import Any, AsyncGenerator, Awaitable, Callable, Dict, Optional, TypeVar

from classiq._internals.authentication.auth0 import Auth0, Tokens
from classiq.exceptions import ClassiqAuthenticationError, ClassiqExpiredTokenError

T = TypeVar("T")


async def _poll_with_timeout(
    timeout: float, callable_: Callable[..., Awaitable[T]], *args, **kwargs
) -> AsyncGenerator[T, None]:
    polling_start = time.perf_counter()
    while time.perf_counter() - polling_start <= timeout:
        yield await callable_(*args, **kwargs)


class DeviceRegistrar:
    _TIMEOUT_ERROR = (
        "Device registration timed out. Please re-initiate the flow and "
        "authorize the device within the timeout."
    )
    _TIMEOUT_SEC: float = timedelta(minutes=15).total_seconds()

    @classmethod
    async def register(cls, get_refresh_token: bool = True) -> Tokens:
        data: Dict[str, Any] = await Auth0.get_device_data(
            get_refresh_token=get_refresh_token
        )

        print(f"Your user code: {data['user_code']}")
        verification_url = data["verification_uri_complete"]
        print(
            f"If a browser doesn't automatically open, please visit the url: {verification_url}"
        )
        webbrowser.open(verification_url)
        timeout = min(data["expires_in"], cls._TIMEOUT_SEC)
        return await cls._poll_tokens(
            device_code=data["device_code"],
            interval=data["interval"],
            timeout=timeout,
            get_refresh_token=get_refresh_token,
        )

    @classmethod
    def _handle_ready_data(
        cls, data: Dict[str, Any], get_refresh_token: bool
    ) -> Tokens:
        access_token: Optional[str] = data.get("access_token")
        # If refresh token was not requested, this would be None
        refresh_token: Optional[str] = data.get("refresh_token")

        if access_token is None or (
            get_refresh_token is True and refresh_token is None
        ):
            raise ClassiqAuthenticationError(
                "Token generation failed for unknown reason."
            )

        return Tokens(access_token=access_token, refresh_token=refresh_token)

    @classmethod
    async def _poll_tokens(
        cls,
        device_code: str,
        interval: int,
        timeout: int,
        get_refresh_token: bool = True,
    ) -> Tokens:
        await asyncio.sleep(interval)
        async for data in _poll_with_timeout(
            timeout, Auth0.poll_tokens, device_code=device_code
        ):
            error_code: Optional[str] = data.get("error")
            if error_code is None:
                return cls._handle_ready_data(data, get_refresh_token)
            elif error_code == "authorization_pending":
                pass
            elif error_code == "slow_down":
                interval *= 2
            elif error_code == "expired_token":
                raise ClassiqExpiredTokenError(cls._TIMEOUT_ERROR)
            elif error_code == "access_denied":
                raise ClassiqAuthenticationError("Access denied.")
            else:
                raise ClassiqAuthenticationError(
                    f"Device registration failed with an unknown error: {error_code}."
                )
            await asyncio.sleep(interval)
        else:
            raise ClassiqAuthenticationError(cls._TIMEOUT_ERROR)
